package main

import (
	"errors"
	"fmt"
	"io/ioutil"
	"log"
	"net/http"
	"os"
	"time"
	"encoding/json"

	go2sky "github.com/SkyAPM/go2sky"
	httpPlugin "github.com/SkyAPM/go2sky/plugins/http"
	"github.com/SkyAPM/go2sky/reporter"
)

const (
	registerURL      = "/extension/register"
	invokeURL        = "/extension/invoke"
	requestTimeout   = 5 * time.Second
	ExtensionApiAddr = "EXTENSION_API_ADDR"
	ExtensionName    = "FunctionGraph-Extension-Name"
	ExtensionAddr    = "FunctionGraph-Extension-Address"

	currentExtensionAddr = "127.0.0.1:8081"
	currentExtensionName = "extension_1.sh"
)

// Configuration -
type ExtensionRequest struct {
	TraceID string `json:"traceId" valid:"optional"`
	URN     string `json:"invokedFunctionUrn" valid:"optional"`
}

func main() {
	liveChan := make(chan struct{})
	go startSkywalkingAgent()
	err := registerExtension(currentExtensionName, currentExtensionAddr)
	if err != nil {
		fmt.Println(" failed to register extension, error: %s ", err.Error())
		return
	}
	<-liveChan
}

func startSkywalkingAgent() {
	r, err := reporter.NewGRPCReporter("10.174.242.117:11800", reporter.WithCDS(100))
	if err != nil {
		log.Fatalf("new reporter error %v \n", err)
		return
	}
	defer r.Close()
	tracer, err := go2sky.NewTracer("testservice", go2sky.WithReporter(r))
	if err != nil {
		log.Fatalf("create tracer error %v \n", err)
		return
	}
	go2sky.SetGlobalTracer(tracer)
	sm, err := httpPlugin.NewServerMiddleware(go2sky.GetGlobalTracer())
	if err != nil {
		log.Fatalf("create server middleware error %v \n", err)
		return
	}
	http.HandleFunc(invokeURL, sm(RequestHandler{}).ServeHTTP)
	http.ListenAndServe(currentExtensionAddr, nil)
}

func getHTTPClient() *http.Client {
	httpClient := &http.Client{
		Timeout: requestTimeout,
	}
	return httpClient
}

func registerExtension(name, address string) error {
	addr := os.Getenv("EXTENSION_API_ADDR")
	if len(addr) == 0 {
		return errors.New("failed to get extension api addr")
	}
	baseURLPath := "http://" + addr + registerURL
	req, err := http.NewRequest("POST", baseURLPath, nil)
	if err != nil {
		fmt.Printf("failed to build register request, error: %s\n", err.Error())
		return err
	}
	req.Header.Set("Content-Type", "application/json")
	req.Header.Set(ExtensionName, currentExtensionName)
	req.Header.Set(ExtensionAddr, currentExtensionAddr)
	_, err = getHTTPClient().Do(req)
	if err != nil {
		fmt.Printf("failed to send register request, error: %s\n", err.Error())
		return err
	}
	return nil
}

type RequestHandler struct {
}

func (r RequestHandler) ServeHTTP(writer http.ResponseWriter, request *http.Request) {
	req := ExtensionRequest{}
	defer request.Body.Close()
	b, err := ioutil.ReadAll(request.Body)
	if err != nil {
		fmt.Printf("failed to parse invoke request body with error: %s", err.Error())
		writeHeaders(writer, 500, "failed to parse invoke request body")
		return
	}
	if err := json.Unmarshal(b, &req); err != nil {
		fmt.Printf("failed to unmarshal the json, error: %s", err.Error())
		writeHeaders(writer, 500, "failed to unmarshal the json")
		return
	}

	fmt.Println(" invoke request traceID -- ", req.TraceID)
	fmt.Println(" invoke request URN -- ", req.URN)
	writeHeaders(writer, 200, "succeed to get invoke request")
	return
}

func writeHeaders(w http.ResponseWriter, status int, msg string) error {
	w.WriteHeader(status)
	_, err := w.Write([]byte(msg))
	return err
}
